﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.EvolutionTracking {
  [Item("DirectedGraph", "Generic class representing a directed graph with custom vertices and content")]
  [StorableClass]
  public class DirectedGraph : Item, IDirectedGraph {
    public bool AllowDuplicateContent { get; set; }

    [Storable]
    private readonly List<IVertex> nodes; // for performance reasons, maybe this should be a linked list (fast remove)
    public IEnumerable<IVertex> Nodes {
      get { return nodes; }
    }

    [Storable]
    private readonly Dictionary<string, IVertex> idMap;

    [Storable]
    private readonly Dictionary<object, IVertex> contentMap;

    [StorableConstructor]
    protected DirectedGraph(bool serializing)
      : base(serializing) {
    }

    public DirectedGraph() {
      nodes = new List<IVertex>();
      contentMap = new Dictionary<object, IVertex>();
      idMap = new Dictionary<string, IVertex>();
    }

    protected DirectedGraph(DirectedGraph original, Cloner cloner)
      : base(original, cloner) {
      nodes = new List<IVertex>(original.Nodes);
      contentMap = new Dictionary<object, IVertex>(original.contentMap);
      idMap = new Dictionary<string, IVertex>(original.idMap);
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new DirectedGraph(this, cloner);
    }

    public bool Contains(IVertex t) {
      return nodes.Contains(t);
    }

    public bool Contains(object content) {
      return contentMap.ContainsKey(content);
    }

    public IVertex GetVertex(string id) {
      IVertex result;
      idMap.TryGetValue(id, out result);
      return result;
    }

    public IVertex GetVertex(object content) {
      IVertex result;
      contentMap.TryGetValue(content, out result);
      return result;
    }

    public virtual bool Any(Func<IVertex, bool> predicate) {
      return nodes.Any(predicate);
    }

    public virtual bool IsEmpty {
      get { return !nodes.Any(); }
    }

    public virtual void Clear() {
      nodes.Clear();
      contentMap.Clear();
      idMap.Clear();
    }

    public virtual void AddVertex(IVertex vertex) {
      if (vertex.Content != null) {
        if (!AllowDuplicateContent && contentMap.ContainsKey(vertex.Content)) {
          throw new Exception("Content already exists in the graph.");
        }
        contentMap.Add(vertex.Content, vertex);
      }
      idMap.Add(vertex.Id, vertex);
      nodes.Add(vertex);

      vertex.PreContentChanged += Vertex_PreContentChanged;
      vertex.PostContentChanged += Vertex_PostContentChanged;
    }

    public virtual void RemoveVertex(IVertex vertex) {
      contentMap.Remove(vertex.Content);
      idMap.Remove(vertex.Id);
      nodes.Remove(vertex);

      vertex.PreContentChanged -= Vertex_PreContentChanged;
      vertex.PostContentChanged -= Vertex_PostContentChanged;
    }

    public virtual IArc AddArc(IVertex source, IVertex target) {
      var arc = new Arc(source, target);
      source.AddForwardArc(arc);
      target.AddReverseArc(arc);
      return arc;
    }

    private void Vertex_PreContentChanged(object sender, EventArgs args) {
      var vertex = (IVertex)sender;
      if (contentMap.ContainsKey(vertex.Content)) {
        contentMap.Remove(vertex.Content);
      }
    }

    private void Vertex_PostContentChanged(object sender, EventArgs args) {
      var vertex = (IVertex)sender;
      if (vertex.Content != null)
        contentMap.Add(vertex.Content, vertex);
    }

    public override Image ItemImage {
      get { return Common.Resources.VSImageLibrary.Graph; }
    }
  }
}
